/* BlastZ binary alignment visualization tool
 * Requires gd library available at www.boutell.com
 * Ivan Ovcharenko <ovcharenko1@llnl.gov>
 */

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "ctype.h"
#include "zPicture.h"


int main(int argc, char *argv[]) {
    unsigned long int maxChromSize(0);
    char *BinAlnFileName, *Seq1FileName, *Seq2FileName, *ECRFileName=NULL, *ExonsFileName, *RepeatsFileName;

    // Default parameters
    ecrWin = 100;
    ecrId = 70;
    nLayerBases = 25000;
    xMax = 850;
    yLayer = 120;
    percFrom = 50;
    percTo = 100;
    topBottom = 0;
    to = from = 0;
    removeLegend = 0;
    justPrintEcrs = 0;
    calculateId = 0;
    nGenes = nAnnoLines;
    Gene = NULL;
    Annotation = NULL;
    ECR = NULL;
    PipConservation = Conservation = NULL;
    SMOOTH = 0;
    numSeq = 2;

    // extract command line options
    getOptions(argc,argv);

    // consistency check point
    if ((calculateId==1) && ((from==0) || (to==0))) {
	    fprintf(stderr, "Error. -from and -to options have to be utilized together with -calculateId option\n");
	    exit(1);
    }

    //printf("nLayerBases: %d\nxMax: %d\nyLayer: %d\n",nLayerBases,xMax,yLayer);
    
    // extract command line parameters
    if ( argc != (6+(numSeq-2)*2) ) {
	fprintf (stderr,"Usage: %s [options] file.bzn seq1.fa seq2.fa file.exons file.reps\n",argv[0]);

	fprintf (stderr,"\t-numSeq 2 : number of sequences in the alignments\n");
	fprintf (stderr,"\tfile.zbn : Binary blastz output file\n");
	fprintf (stderr,"\tseq1.fa and seq2.fa : FASTA sequences\n");
	fprintf (stderr,"\tfile.ecr : List of evolutionary conserved regions\n");
	fprintf (stderr,"\tfile.exons : Gene annotation\n");
	fprintf (stderr,"\tfile.reps : RepeatMasker output\n");
	fprintf (stderr,"\t-ecrWin 100 : Minimum length of an ECR (evolutionary conserved region)\n");
	fprintf (stderr,"\t-ecrId 70 : ECR minimum percent identity\n");
	fprintf (stderr,"\t-layerBases 25000 : Number of bases per layer\n");
	fprintf (stderr,"\t-xMax 850 : Width of the picture\n");
	fprintf (stderr,"\t-yLayer 100 : Height of a layer\n");
	fprintf (stderr,"\t-percFrom 50 -percTo 100 : Starting/Ending position on Y (conservation %) axis\n");
	fprintf (stderr,"\t-justPrintEcrs : print out ECRs, do not output the picture\n");
	fprintf (stderr,"\t-calculateId : calculate percent identity in between given boundaries\n");
	fprintf (stderr,"\t-switch : switch top and bottom sequences\n");
	fprintf (stderr,"\t-smooth : smooth graph instead of a SMOOTH-plot\n");
	fprintf (stderr,"\t-noLegend : remove legend\n");
	exit(1); }
    else {
	GenBinAlnFileName = (char **) calloc( numSeq+1, sizeof(char *) );
	GenSeqFileName = (char **) calloc( numSeq+1, sizeof(char *) );
	for (int iS=2; iS<=numSeq; iS++) 
	    GenBinAlnFileName[iS] = strdup(argv[1+(iS-2)]);
	for (int iS=1; iS<=numSeq; iS++) 
	    GenSeqFileName[iS] = strdup(argv[2 + (numSeq-2) + (iS-1)]);
	ExonsFileName = strdup(argv[ 3 + (numSeq-2) + (numSeq-1) ]);
	RepeatsFileName = strdup(argv[ 4 + (numSeq-2) + (numSeq-1) ]);
    }

    // Initialize numSeq dependent parameters
    seq = (char **) calloc( numSeq+1, sizeof(char *) );
    header1 = (char **) calloc( numSeq+1, sizeof(char *) );
    header2 = (char **) calloc( numSeq+1, sizeof(char *) );
    length1 = (int *) calloc( numSeq+1, sizeof(int) );
    length2 = (int *) calloc( numSeq+1, sizeof(int) );
    nAlignments = (int *) calloc( numSeq+1, sizeof(int) );
    ALT = (alignBlock **) calloc( numSeq+1, sizeof(alignBlock *) );
    Conservation = (int **) calloc( numSeq+1, sizeof(int*) );
    PipConservation = (int **) calloc( numSeq+1, sizeof(int*) );
    ECR = (long int **) calloc( numSeq+1, sizeof(long int*) );
    nECR = (long int *) calloc( numSeq+1, sizeof(long int) );
    for (int iS=0; iS<=numSeq; iS++) {
	seq[iS] = NULL; 
	header1[iS] = NULL; 
	header2[iS] = NULL; 
	length1[iS] = 0; 
	length2[iS] = 0; 
	nAlignments[iS] = 0;
	ALT[iS] = NULL;
	Conservation[iS] = NULL;
	PipConservation[iS] = NULL;
	nECR[iS] = 0;
	ECR[iS] = NULL;
    }

    // 1. READ seq1...seqJ SEQUENCES 
    ReadSeqs();

    // 2. Refine requested picture boundaries, COMPUTE consLen
    if (from<=0) from=1;
    consLen=1;
    if (!topBottom) {
	consLen = strlen(seq[1]);
    } else {
        for (int iS=2; iS<=numSeq; iS++) 
	    if (consLen < strlen(seq[iS]))
		consLen = strlen(seq[iS]); 
    }
    if ((to<=0) || (to>consLen)) to=consLen;
    from = (from < 1) ? 1 : from;
    from = (from > to) ? to : from;

    // 3. READ ALIGNMENTS
    // Read BLASTZ alignments into the ALT[][] structure
    for (int iS=2; iS<=numSeq; iS++) 
    	Get_BLASTZ_Alignment( iS );

    // 4. Consistency check point
    //for (int iS=2; iS<=numSeq; iS++) 
    //	if (strlen(seq[iS])!=length2[iS]) { fprintf(stderr,"Sequence length inconsistency in the alignment file and the fasta files...\n"); exit(1); }
   
    // 5. CREAT CONSERVATION PROFILES
    for (int iS=2; iS<=numSeq; iS++)  {
    	// Reformat seq[1] and seq[iS] sequences --> MATCHES are in UPPER CASE, mismatches (gaps) are in lower case
    	ReformatSeqs( iS );

	if (calculateId==1) {
		calcId( from, to );
	} else {
	        // Create 'Conservation' or 'PipConservation' array 
        	if (SMOOTH) 
		    Get_BLASTZ_Conservation(WINDOW, iS);
        	else 
		    Get_BLASTZ_PipConservation( iS );
    
	        // Determine location of ECRs
	        Define_ECRs( iS );

	        // Join overlapping ECRs
        	JoinECROverlaps( iS );
	}
    }
    if (calculateId==1)
	    exit(0);
    

    // 5. Get Annotation
    // Annotation is a 3*N array containing [start, end, type]
    //     where type = 1 - gene, 2 - exon, 3 - utr, 4 - repeat
    GetAnnotation(ExonsFileName, RepeatsFileName);

    // Either CREATE A PICTURE or OUTPUT ECRs
    if (!justPrintEcrs) 
    	CreatePicture(from, to);
    else 
	for (int iS=2; iS<=numSeq; iS++) 
		PrintEcrs(from, to, iS);
   
    // Free memory
    for (int iS=2; iS<=numSeq; iS++) 
	delete (GenBinAlnFileName[iS]);
    for (int iS=1; iS<=numSeq; iS++) 
	delete (GenSeqFileName[iS]);
    delete GenBinAlnFileName;
    delete GenSeqFileName;
    if (ECRFileName!=NULL)
    	delete ECRFileName;
    delete ExonsFileName;
    delete RepeatsFileName;
    for (int i=1; i<=numSeq; i++) 
        if (seq[i]!=NULL)
	    delete seq[i];
    for (int iGene=0; iGene<nGenes; iGene++) 
	delete Gene[iGene].name;
    if (nGenes>0)
	delete Gene;
}

//////////////////////////////////////////////////////////////////////////////////////////////

int checkOption(int &i, int &argc, char *argv[], char *parameter, int &value, int type) {
    if (type==2) {
	if ((strcmp(argv[i],parameter)==0) && ((i+1)<argc)) {
	    value = atoi(argv[i+1]);
	    for (int j=i; j<argc-2; j++) 
		argv[j] = argv[j+2];
	    argc -= 2;
	    --i;
	    //printf("parameter: %s; value: %d\n",parameter,value);
	    return 1;
	}
    }
    else if (type==1) {
	if (strcmp(argv[i],parameter)==0) {
	    value = 1;
	    for (int j=i; j<argc-1; j++) 
		argv[j] = argv[j+1];
	    --argc;
	    --i;
	    //printf("parameter: %s --> 1\n",parameter);
	    return 1;
	}
    }
    return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////

void getOptions(int &argc, char *argv[]) {
    for (int i=0; i<argc; i++) {
	if (strncmp(argv[i],"-",1)==0) {

	    //processing an option
	    checkOption(i, argc,argv, "-layerBases",nLayerBases, 2)
		||
		checkOption(i, argc,argv, "-numSeq",numSeq, 2)
		||
		checkOption(i, argc,argv, "-xMax",xMax, 2)
		||
		checkOption(i, argc,argv, "-from",from, 2)
		||
		checkOption(i, argc,argv, "-to",to, 2)
		||
		checkOption(i, argc,argv, "-yLayer",yLayer, 2)
		||
		checkOption(i, argc,argv, "-percFrom",percFrom, 2)
		||
		checkOption(i, argc,argv, "-percTo",percTo, 2)
		||
		checkOption(i, argc,argv, "-ecrWin",ecrWin, 2)
		||
		checkOption(i, argc,argv, "-ecrId",ecrId, 2)
		||
		checkOption(i, argc,argv, "-calculateId",calculateId, 1)
		||
		checkOption(i, argc,argv, "-justPrintEcrs",justPrintEcrs, 1)
		||
		checkOption(i, argc,argv, "-switch",topBottom, 1)
		||
		checkOption(i, argc,argv, "-noLegend",removeLegend, 1)
		||
		checkOption(i, argc,argv, "-smooth",SMOOTH, 1);
	}	
    }
    if (removeLegend) xPICFROM=0;
}

//////////////////////////////////////////////////////////////////////////////////////////////

void GetAnnotation(char *ExonsFileName, char *RepeatsFileName) {
    FILE *ExonsFile, *RepeatsFile;
    long int start, end;

    // NULL Genes array
    Gene = NULL;
    nGenes = 0;

    // Open files for reading
    if ((ExonsFile = fopen(ExonsFileName,"r")) == NULL) {
	//printf ("Error. Can not open %s file\n",ExonsFileName);
	//exit(1);
    }
    if ((RepeatsFile = fopen(RepeatsFileName,"r")) == NULL) {
	//printf ("Error. Can not open %s file\n",RepeatsFileName);
	//exit(1);
    }
    
    // Allocate the initial segment of memory 
    if ((Annotation = (long int *) malloc(3*sizeof(long int))) == NULL) {
	printf("Error: Not enough memory for an annotation data storage (malloc)\n");
	exit(1);
    }
    nAnnoLines = 0;

    // Extract exons
    char buffer[1000];
    char type[1000];
    int iType = -1;
    char cDir;
    char gName[1000];
    while ((ExonsFile!=NULL) && (fgets(buffer,1000,ExonsFile) != NULL)) {
	if (sscanf(buffer,"%d %d %s\n",&start,&end,type) != 3) {
	    // not utr or marked exon
	    if (sscanf(buffer,"%[><] %d %d %s\n",&cDir,&start,&end,gName) == 4) {
		iType = 1;
		// Store new gene
		if ((Gene = (GeneStruct *) realloc(Gene,(nGenes+1)*sizeof(GeneStruct))) == NULL) {
		    printf("Error: Not enough memory for a gene annotation data storage (realloc)\n");
		    exit(1);
		}
		if (cDir=='>') Gene[nGenes].strand=1; else Gene[nGenes].strand=-1;
		Gene[nGenes].name = strdup(gName);
		Gene[nGenes].from = start;
		Gene[nGenes].to = end;
		//fprintf(stderr,"%d %d %d %s\n",Gene[nGenes].strand,Gene[nGenes].from,Gene[nGenes].to,Gene[nGenes].name);
		nGenes++;
	    } else if (sscanf(buffer,"%d %d\n",&start,&end) == 2) {
		iType = 2;
	    } else {
		start = -1;
		iType = -1;
	    }
	} else {
	    if (strcasecmp(type,"utr")==0) {
		iType = 3;
	    } else if (strcasecmp(type,"oth")==0) {
		iType = 6;
	    } else {
		iType = 2;
	    }
	}
	if (start > -1) {
	    // Store the data to the Annotation
	    ++nAnnoLines;
	    if ((Annotation = (long int *) realloc(Annotation,3*nAnnoLines*sizeof(long int))) == NULL) {
		printf("Error: Not enough memory for an annotation data storage (realloc)\n");
		exit(1);
	    }
	    Annotation[3*(nAnnoLines-1)+0] = start;
	    Annotation[3*(nAnnoLines-1)+1] = end;
	    Annotation[3*(nAnnoLines-1)+2] = iType;
	}
    }

    // Extract REPEATS
    iType = -1;
    while ((RepeatsFile!=NULL) && (fgets(buffer,1000,RepeatsFile) != NULL)) {
	if (sscanf(buffer," %*d %*f %*f %*f %*s %d %d",&start,&end) == 2) {
	    // Store the data to the Annotation as a REPEAT
	    ++nAnnoLines;
	    if ((Annotation = (long int *) realloc(Annotation,3*nAnnoLines*sizeof(long int))) == NULL) {
		printf("Error: Not enough memory for an annotation data storage (realloc)\n");
		exit(1);
	    }
	    Annotation[3*(nAnnoLines-1)+0] = start;
	    Annotation[3*(nAnnoLines-1)+1] = end;
	    Annotation[3*(nAnnoLines-1)+2] = 4;
	}
    }
    
    // Consensus check
    //int i;
    //for (i=0; i<nAnnoLines; i++) 
    //	printf ("%d (%d):\t%d-%d\t%d \n", i, nAnnoLines, Annotation[i*3+0], Annotation[i*3+1], Annotation[i*3+2]);
    
    // Close files
    if (ExonsFile!=NULL)
	fclose(ExonsFile);
    if (RepeatsFile!=NULL)
	fclose(RepeatsFile);
}

//////////////////////////////////////////////////////////////////////////////////////////////


